class alttprException(Exception):
    pass


class alttprFailedToRetrieve(Exception):
    pass


class alttprFailedToGenerate(Exception):
    pass


class UnableToRetrieve(Exception):
    pass

class UnableToGenerate(Exception):
    pass


from . import misc
from .exceptions import alttprException

import hashlib
import aiofiles


async def read(srcfilepath, verify_checksum=True):
    """Reads a ROM and converts it to a list of bytes, preparing it for patching.

    Arguments:
        srcfilepath {str} -- A path (either relative or absolute) to a file on the filesystem to be read.
        verify_checksum {bool} -- Verify the checksum of the file to ensure it is an unheadered ALTTPR Japan 1.0 ROM.

    Raises:
        alttprException -- Returns if the sha256 checksum does not match a ALTTPR Japan 1.0 ROM.

    Returns:
        list -- A list of bytes depicting the read ROM file.
    """
    async with aiofiles.open(srcfilepath, "rb") as f:
        baserom_array = list(await f.read())
    if verify_checksum:
        if len(baserom_array) == 1049088:
            baserom_array = baserom_array[512:]
        expected_rom_sha256 = '794e040b02c7591b59ad8843b51e7c619b88f87cddc6083a8e7a4027b96a2271'
        sha256_hash = hashlib.sha256()
        sha256_hash.update(bytes(baserom_array))
        if not sha256_hash.hexdigest() == expected_rom_sha256:
            raise alttprException(
                'Expected checksum "{expected_rom_sha256}", got "{actual_checksum}" instead.  Verify the source ROM is an unheadered Japan 1.0 Link to the Past ROM.'.format(
                    expected_rom_sha256=expected_rom_sha256, actual_checksum=sha256_hash.hexdigest()))
    return baserom_array

async def write(rom, dstfilepath):
    fw = await aiofiles.open(dstfilepath, "wb")
    patchrom = bytes()
    for idx, chunk_array in enumerate(misc.chunk(rom, 256)):
        patchrom += bytes(chunk_array)
    await fw.write(patchrom)
    fw.close



###
# Originaly created by Hycutype
###


# Priestmode
# Triforce is in sanctuary chest, goal is to beat enemized escape
# quickswap is allowed
# thieves are killable
# the sewers key may be in the uncle passage
#
# bombs are guaranteed as drops in case the randomized weapon cannot kill the
# boomerang chest guard or the keyrat (gibdo or minihelmasaur)
#   * prize packs 4 & 7 are unmodified; they each have an 8bomb drop
#   * the second prize of all other packs is replaced with a 4bomb drop
#   * the _secondary_ crab prize is a 4bomb drop

# technical:
# goal is triforce-hunt, suggested by Synack
# prize packs are manually shuffled. Synack confirmed this is necessary with
#   custom packs
# it's smallkey shuffle, with every key besides the sewers key manually
#   placed. I guessed this would be least problematic for the customizer
# key in link's house shouldn't be possible, because once you collect the
#   uncle item you can't get back

# observations:
# I was worried the extra bombs would be too much, but it's possible they are
# actually inadequate; we'll see with more plays. I am tempted to mess with
# the item pool more, or do more manual placments to give a more even
# distribution across seeds in a batch (e.g. the uncle weapon, weighting for
# weapons that haven't been placed in earlier seeds)
import json
import random
# import urllib.request as request
# from threading import Thread
from copy import deepcopy
from math import factorial
from pyz3r.alttpr import alttprClass as alttpr

ORIG_SETTINGS = {
    "allow_quickswap": True,
    "glitches": "none",
    "item_placement": "advanced",
    "dungeon_items": "standard",
    "accessibility": "none",
    "goal": "triforce-hunt",
    "crystals": {
        "ganon": "7",
        "tower": "7"
    },
    "mode": "standard",
    "hints": "off",
    "weapons": "randomized",
    "item": {
        "pool": "normal",
        "functionality": "normal"
    },
    "tournament": True,
    "spoilers": "off",
    "lang": "en",
    "enemizer": {
        "boss_shuffle": "none",
        "enemy_shuffle": "shuffled",
        "enemy_damage": "default",
        "enemy_health": "default"
    },
    "name": "kiss-priest",
    "notes": "The Triforce is in sanc chest. THE KEY MAY BE IN THE UNCLE CHEST\n\nBombs are guaranteed as drops in case the uncle weapon can't kill the boomerang chest guard or keyrat. Packs 4 & 7 are unmodified (8-bombs). The second prize of all others is replaced with a 4-bomb. The second crab drop is a 4-bomb",
    "l": {
        "U2FuY3R1YXJ5OjE=": "Triforce:1",
        "RGVzZXJ0IFBhbGFjZSAtIFRvcmNoOjE=": "KeyP2:1",
        "VG93ZXIgb2YgSGVyYSAtIEJhc2VtZW50IENhZ2U6MQ==": "KeyP3:1",
        "Q2FzdGxlIFRvd2VyIC0gUm9vbSAwMzox": "KeyA1:1",
        "Q2FzdGxlIFRvd2VyIC0gRGFyayBNYXplOjE=": "KeyA1:1",
        "UGFsYWNlIG9mIERhcmtuZXNzIC0gU2hvb3RlciBSb29tOjE=": "KeyD1:1",
        "UGFsYWNlIG9mIERhcmtuZXNzIC0gVGhlIEFyZW5hIC0gTGVkZ2U6MQ==": "KeyD1:1",
        "UGFsYWNlIG9mIERhcmtuZXNzIC0gVGhlIEFyZW5hIC0gQnJpZGdlOjE=": "KeyD1:1",
        "UGFsYWNlIG9mIERhcmtuZXNzIC0gU3RhbGZvcyBCYXNlbWVudDox": "KeyD1:1",
        "UGFsYWNlIG9mIERhcmtuZXNzIC0gRGFyayBCYXNlbWVudCAtIExlZnQ6MQ==": "KeyD1:1",
        "UGFsYWNlIG9mIERhcmtuZXNzIC0gRGFyayBCYXNlbWVudCAtIFJpZ2h0OjE=": "KeyD1:1",
        "U3dhbXAgUGFsYWNlIC0gRW50cmFuY2U6MQ==": "KeyD2:1",
        "U2t1bGwgV29vZHMgLSBCcmlkZ2UgUm9vbTox": "KeyD3:1",
        "U2t1bGwgV29vZHMgLSBQb3QgUHJpc29uOjE=": "KeyD3:1",
        "U2t1bGwgV29vZHMgLSBQaW5iYWxsIFJvb206MQ==": "KeyD3:1",
        "VGhpZXZlcycgVG93biAtIEFtYnVzaCBDaGVzdDox": "KeyD4:1",
        "SWNlIFBhbGFjZSAtIEZyZWV6b3IgQ2hlc3Q6MQ==": "KeyD5:1",
        "SWNlIFBhbGFjZSAtIEljZWQgVCBSb29tOjE=": "KeyD5:1",
        "TWlzZXJ5IE1pcmUgLSBNYWluIExvYmJ5OjE=": "KeyD6:1",
        "TWlzZXJ5IE1pcmUgLSBCcmlkZ2UgQ2hlc3Q6MQ==": "KeyD6:1",
        "TWlzZXJ5IE1pcmUgLSBTcGlrZSBDaGVzdDox": "KeyD6:1",
        "VHVydGxlIFJvY2sgLSBDaGFpbiBDaG9tcHM6MQ==": "KeyD7:1",
        "VHVydGxlIFJvY2sgLSBSb2xsZXIgUm9vbSAtIExlZnQ6MQ==": "KeyD7:1",
        "VHVydGxlIFJvY2sgLSBSb2xsZXIgUm9vbSAtIFJpZ2h0OjE=": "KeyD7:1",
        "VHVydGxlIFJvY2sgLSBCaWcgQ2hlc3Q6MQ==": "KeyD7:1",
        "R2Fub24ncyBUb3dlciAtIEJvYidzIFRvcmNoOjE=": "KeyA2:1",
        "R2Fub24ncyBUb3dlciAtIERNcyBSb29tIC0gVG9wIExlZnQ6MQ==": "KeyA2:1",
        "R2Fub24ncyBUb3dlciAtIERNcyBSb29tIC0gVG9wIFJpZ2h0OjE=": "KeyA2:1",
        "R2Fub24ncyBUb3dlciAtIERNcyBSb29tIC0gQm90dG9tIExlZnQ6MQ==": "KeyA2:1"
    },
    "eq": ["BossHeartContainer", "BossHeartContainer", "BossHeartContainer"],
    "drops": {
        "0": ["Heart", "BombRefill4", "Heart", "Heart", "RupeeGreen", "Heart", "Heart", "RupeeGreen"],
        "1": ["RupeeBlue", "BombRefill4", "RupeeBlue", "RupeeRed", "RupeeBlue", "RupeeGreen", "RupeeBlue", "RupeeBlue"],
        "2": ["MagicRefillFull", "BombRefill4", "MagicRefillSmall", "RupeeBlue", "MagicRefillFull", "MagicRefillSmall", "Heart", "MagicRefillSmall"],
        "3": ["BombRefill1", "BombRefill1", "BombRefill1", "BombRefill4", "BombRefill1", "BombRefill1", "BombRefill8", "BombRefill1"],
        "4": ["ArrowRefill5", "BombRefill4", "ArrowRefill5", "ArrowRefill10", "ArrowRefill5", "Heart", "ArrowRefill5", "ArrowRefill10"],
        "5": ["MagicRefillSmall", "BombRefill4", "Heart", "ArrowRefill5", "MagicRefillSmall", "BombRefill1", "RupeeGreen", "Heart"],
        "6": ["Heart", "Fairy", "MagicRefillFull", "RupeeRed", "BombRefill8", "Heart", "RupeeRed", "ArrowRefill10"],
        "pull": ["RupeeGreen", "RupeeGreen", "RupeeGreen"],
        "crab": ["auto_fill", "BombRefill4"],
        "stun": ["auto_fill"],
        "fish": ["RupeeGreen"]
    },
    "custom": {
        "item.Goal.Required": "",
        "item.require.Lamp": False,
        "item.value.BlueClock": "",
        "item.value.GreenClock": "",
        "item.value.RedClock": "",
        "item.value.Rupoor": "",
        "prize.crossWorld": True,
        "prize.shuffleCrystals": True,
        "prize.shufflePendants": True,
        "region.bossNormalLocation": True,
        "region.wildBigKeys": False,
        "region.wildCompasses": False,
        "region.wildKeys": True,
        "region.wildMaps": False,
        "rom.dungeonCount": "off",
        "rom.freeItemMenu": False,
        "rom.freeItemText": False,
        "rom.mapOnPickup": False,
        "rom.timerMode": "off",
        "rom.timerStart": "",
        "rom.rupeeBow": False,
        "rom.genericKeys": False,
        "rom.logicMode": "NoGlitches",
        "spoil.BootsLocation": False,
        "canBootsClip": False,
        "canBunnyRevive": False,
        "canBunnySurf": False,
        "canDungeonRevive": False,
        "canFakeFlipper": False,
        "canMirrorClip": False,
        "canMirrorWrap": False,
        "canOneFrameClipOW": False,
        "canOneFrameClipUW": False,
        "canOWYBA": False,
        "canSuperBunny": False,
        "canSuperSpeed": False,
        "canWaterWalk": False,
        "item": {
            "count": {
                "BottleWithRandom": 4,
                "Nothing": 0,
                "L1Sword": 0,
                "L1SwordAndShield": 0,
                "MasterSword": 0,
                "L3Sword": 0,
                "L4Sword": 0,
                "BlueShield": 0,
                "RedShield": 0,
                "MirrorShield": 0,
                "FireRod": 1,
                "IceRod": 1,
                "Hammer": 1,
                "Hookshot": 1,
                "Bow": 0,
                "Boomerang": 1,
                "Powder": 1,
                "Bombos": 1,
                "Ether": 1,
                "Quake": 1,
                "Lamp": 1,
                "Shovel": 1,
                "OcarinaInactive": 1,
                "CaneOfSomaria": 1,
                "Bottle": 0,
                "PieceOfHeart": 24,
                "CaneOfByrna": 1,
                "Cape": 1,
                "MagicMirror": 1,
                "PowerGlove": 0,
                "TitansMitt": 0,
                "BookOfMudora": 1,
                "Flippers": 1,
                "MoonPearl": 1,
                "BugCatchingNet": 1,
                "BlueMail": 0,
                "RedMail": 0,
                "Bomb": 0,
                "ThreeBombs": 16,
                "Mushroom": 1,
                "RedBoomerang": 1,

                "BottleWithGreenPotion": 0,
                "BottleWithBluePotion": 0,
                "TenBombs": 1,
                "OneRupee": 2,
                "FiveRupees": 4,
                "TwentyRupees": 27,
                "BowAndArrows": 0,
                "BowAndSilverArrows": 0,
                "BottleWithBee": 0,
                "BottleWithFairy": 0,
                "BossHeartContainer": 10,
                "HeartContainer": 1,
                "OneHundredRupees": 1,
                "FiftyRupees": 7,
                "Heart": 0,
                "Arrow": 1,
                "TenArrows": 12,
                "SmallMagic": 0,
                "ThreeHundredRupees": 5,
                "BottleWithGoldBee": 0,
                "OcarinaActive": 0,
                "PegasusBoots": 1,
                "BombUpgrade5": 0,
                "BombUpgrade10": 0,
                "ArrowUpgrade5": 0,
                "ArrowUpgrade10": 0,
                "HalfMagic": 1,
                "QuarterMagic": 0,
                "SilverArrowUpgrade": 0,
                "Rupoor": 0,
                "RedClock": 0,
                "BlueClock": 0,
                "GreenClock": 0,
                "ProgressiveSword": 4,
                "ProgressiveShield": 3,
                "ProgressiveArmor": 2,
                "ProgressiveGlove": 2,
                "ProgressiveBow": 2,
                "Triforce": 0,
                "TriforcePiece": 0,
                "MapA2": 1,
                "MapD7": 1,
                "MapD4": 1,
                "MapP3": 1,
                "MapD5": 1,
                "MapD3": 1,
                "MapD6": 1,
                "MapD1": 1,
                "MapD2": 1,
                "MapA1": 0,
                "MapP2": 1,
                "MapP1": 1,
                "MapH1": 0,
                "MapH2": 1,
                "CompassA2": 1,
                "CompassD7": 1,
                "CompassD4": 1,
                "CompassP3": 1,
                "CompassD5": 1,
                "CompassD3": 1,
                "CompassD6": 1,
                "CompassD1": 1,
                "CompassD2": 1,
                "CompassA1": 0,
                "CompassP2": 1,
                "CompassP1": 1,
                "CompassH1": 0,
                "CompassH2": 0,
                "BigKeyA2": 1,
                "BigKeyD7": 1,
                "BigKeyD4": 1,
                "BigKeyP3": 1,
                "BigKeyD5": 1,
                "BigKeyD3": 1,
                "BigKeyD6": 1,
                "BigKeyD1": 1,
                "BigKeyD2": 1,
                "BigKeyA1": 0,
                "BigKeyP2": 1,
                "BigKeyP1": 1,
                "BigKeyH1": 0,
                "BigKeyH2": 0,
                "KeyH2": 1,
                "KeyH1": 0,
                "KeyP1": 0,
                "KeyP2": 0,
                "KeyA1": 0,
                "KeyD2": 0,
                "KeyD1": 0,
                "KeyD6": 0,
                "KeyD3": 0,
                "KeyD5": 0,
                "KeyP3": 0,
                "KeyD4": 0,
                "KeyD7": 0,
                "KeyA2": 0
            }
        },
        "drop": {
            "count": {
                "Bee": 0,
                "BeeGood": 0,
                "Heart": 13,
                "RupeeGreen": 9,
                "RupeeBlue": 7,
                "RupeeRed": 6,
                "BombRefill1": 7,
                "BombRefill4": 1,
                "BombRefill8": 2,
                "MagicRefillSmall": 6,
                "MagicRefillFull": 3,
                "ArrowRefill5": 5,
                "ArrowRefill10": 3,
                "Fairy": 1
            }
        }
    }
}

UNCLE_LOCATION = "TGluaydzIFVuY2xlOjE="


async def create_priestmode(count, genclass=alttpr):
    name = 'kiss-priest-' + batch_code() + '-{:0>2d}'
    settings_lists = [None]*count
    uncle = uncle_weapons(count)

    for i in range(count):
        settings = deepcopy(ORIG_SETTINGS)
        randomize_drops(settings)
        settings["name"] = name.format(i+1)
        settings["l"][UNCLE_LOCATION] = uncle[i]
        settings["custom"]["item"]["count"][uncle[i]] -= 1
        settings_lists[i] = settings

    for i in range(count * 2 // 3, count):
        settings_lists[i]["custom"]["item"]["count"]["BossHeartContainer"] -= 6
        settings_lists[i]["custom"]["item"]["count"]["ProgressiveArmor"] -= 1
        settings_lists[i]["custom"]["item"]["count"]["ProgressiveSword"] -= 2
        settings_lists[i]["custom"]["item"]["count"]["PieceOfHeart"] += 9

    seeds = []
    for i, s in enumerate(settings_lists):
        seed = await genclass.create(settings=settings_lists[i], customizer=True)
        seeds.append(seed)

    return seeds

# TODO will return garbage on batch sizes > 24
# XXX ^ doesn't actually do that and I haven't look at it

# TODO reconsider how I was doing sword weights


def uncle_weapons(count):
    sword = "ProgressiveSword"
    weapons = [sword, "ProgressiveBow", "FireRod", "Hammer",
               "CaneOfByrna", "CaneOfSomaria", "TenBombs"]

    # sword_weights = [82951, 15651, 1329, 67, 2]
    # sword_weights = {count:sword_weights[count] for count in range(len(sword_weights))}
    # sword_count = select_weighted_key(sword_weights)

    weights = {weapons[i]: 8 for i in range(len(weapons))}
    weights[sword] = 1

    uncle = []
    for _ in range(count):
        w = select_weighted_key(weights)
        weights[w] //= 2
        uncle.append(w)

    # make the potential sword more likely earlier, I think
    uncle.reverse()

    return uncle


def select_weighted_key(d):
    total = sum(d.values())

    i = random.randrange(total)
    for key, weight in d.items():
        if i <= weight:
            return key
        i -= weight


def randomize_drops(settings):
    drops = settings["custom"]["drop"]["count"]
    total_drops = sum(drops.values())

    def random_drop():
        item = select_weighted_key(drops)
        drops[item] -= 1
        return item

    settings["drops"]["crab"][0] = random_drop()
    settings["drops"]["stun"][0] = random_drop()

    packs = [value for key, value in settings["drops"].items()
             if key in map(str, range(6))]
    random.shuffle(packs)

    # ensure an 8bomb in the gibdo pack in case of double gibdos
    if "BombRefill4" == packs[2][1]:
        packs[2][1] = "BombRefill8"

    for i in range(6):
        settings["drops"][str(i)] = packs[i]


def batch_code():
    def rand_char():
        return chr(random.randint(ord('A'), ord('Z')))

    return ''.join([rand_char() for _ in range(3)])



import bps.apply
import bps.io
import io

import itertools
from .exceptions import alttprException


def apply(rom, patches):
    """Applies a patch, which is a list of dictionaries

    Arguments:
        rom {list} -- A list of bytes depicting the ROM data to be patched.
        patches {list} -- A list of dictionaries that depict of set of patches to be applied to the ROM.

    Returns:
        list -- a list of bytes depicting the patched rom
    """
    for patch in patches:
        offset = int(list(patch.keys())[0])
        patch_values = list(patch.values())[0]
        for idx, value in enumerate(patch_values):
            rom[offset + idx] = value
    return rom

def apply_bps(rom, patches: bytes):
    """Applies a patch, which is a bps

    Arguments:
        rom {list} -- A list of bytes depicting the ROM data to be patched.
        patches {bytes} -- A series of bytes representing a bps patch

    returns:
        list -- a list of bytes depicting the patched rom
    """
    # create "files" for apply_to_files
    # although there is an apply_to_bytearrays it does not perform CRC
    # validation of input or output, apply_to_files does.
    patch_file = io.BytesIO(patches)
    source_file = io.BytesIO(bytes(rom))
    target_file = io.BytesIO()

    bps.apply.apply_to_files(patch_file, source_file, target_file)
    target_file.seek(0)
    return list(target_file.read())

def heart_speed(speed='half'):
    """Set the low-health warning beep interval.

    Keyword Arguments:
        speed {str} -- Chose the speed at which the low health warning beeps.
            Options are 'off', 'double', 'normal', 'half', and 'quarter'. (default: {'half'})

    Returns:
        list -- a list of dictionaries indicating which ROM address offsets to write and what to write to them
    """
    if speed is None:
        speed = 'normal'
    sbyte = {
        'off': 0,
        'double': 16,
        'normal': 32,
        'half': 64,
        'quarter': 128,
    }
    patch = [{
        '1572915': [sbyte[speed]]
    }]
    return patch

def heart_color(color='red'):
    """Set the color of the hearts on the player's HUD.

    Keyword Arguments:
        color {str} -- The heart color.  Options are 'red', 'blue', 'green', and 'yellow' (default: {'red'})

    Returns:
        list -- a list of dictionaries indicating which ROM address offsets to write and what to write to them
    """

    if color is None:
        color = 'red'
    cbyte = {
        'blue': [44, 13],
        'green': [60, 25],
        'yellow': [40, 9],
        'red': [36, 5],
    }
    byte = cbyte[color][0]
    file_byte = cbyte[color][1]
    patch = [
        {'457246': [byte]},
        {'457248': [byte]},
        {'457250': [byte]},
        {'457252': [byte]},
        {'457254': [byte]},
        {'457256': [byte]},
        {'457258': [byte]},
        {'457260': [byte]},
        {'457262': [byte]},
        {'457264': [byte]},
        {'415073': [file_byte]},
    ]
    return patch

def music(music=True):
    """Enables, or disables, the in-game music.  Useful if you want to use an MSU-1 soundtrack instead.

    Keyword Arguments:
        music {bool} -- If true, music is enabled.  If false, the music id disabled. (default: {True})

    Returns:
        list -- a list of dictionaries indicating which ROM address offsets to write and what to write to them
    """

    return [{'1573402': [0 if music else 1]}]


def quickswap(quickswap=False):
    return [{'1572939': [1 if quickswap else 0]}]


def menu_speed(speed='normal'):
    if speed is None:
        speed = 'normal'
    sbyte = {
        'instant': 0xE8,
        'fast': 0x10,
        'normal': 0x08,
        'slow': 0x04
    }
    patch = [
        {'1572936': [sbyte[speed]]},
        {'449946': [0x20 if menu_speed == 'instant' else 0x11]},
        {'450346': [0x20 if menu_speed == 'instant' else 0x12]},
        {'450793': [0x20 if menu_speed == 'instant' else 0x12]}
        ]
    return patch

def sprite(spr):
    """Creates a patch for to replace Link's sprite with the contents of a XSPR or SPR file.

    Arguments:
        spr {list} -- a list of bytes that depicts a ZSPR or SPR file

    Returns:
        list -- a list of dictionaries indicating which ROM address offsets to write and what to write to them
    """

    if spr[:4] == [90, 83, 80, 82]:
        # stolen from VT's code
        gfx_offset = spr[12] << 24 | spr[11] << 16 | spr[10] << 8 | spr[9]
        palette_offset = spr[18] << 24 | spr[17] << 16 | spr[16] << 8 | spr[15]
        patch = [
            {'524288': spr[gfx_offset:gfx_offset + 28671]},
            {'905992': spr[palette_offset:palette_offset + 120]},
            {'912885': spr[palette_offset + 120:palette_offset + 120 + 3]}
        ]
    # Else treat it like a SPR file instead
    else:
        patch = [
            {'524288': spr[0:28671]},
            {'905992': spr[28672:28791]},
            {
                '912885': [
                    spr[28726],
                    spr[28727],
                    spr[28756],
                    spr[28757],
                ]
            }
        ]
    return patch

def checksum(rom):
    """Writes a patch that fixes a ROM's checksum.  This should be the last patch applied to a ROM before it is written.

    Arguments:
        rom {list} -- a list of bytes depicitng the rom

    Returns:
            list -- a list of dictionaries indicating which ROM address offsets to write and what to write to them
    """

    sum_of_bytes = sum(rom[:32731]) + sum(rom[32736:])
    checksum = (sum_of_bytes + 510) & 65535
    inverse = checksum ^ 65535
    patch = [
        {
            '32732': [
                inverse & 255,
                inverse >> 8,
                checksum & 255,
                checksum >> 8,
            ]
        }
    ]
    return patch

def expand(rom, newlenmb):
    """Expands the byte list of a ROM to the specified number of megabytes, filling in the new space with zeroes.

    Arguments:
        rom {list} -- a list of bytes depicitng the rom

    Keyword Arguments:
        newlenmb {int} -- The size of the ROM should be, in megabytes.

    Raises:
        alttprException -- Raised if the new length is shorter than the current size of the byte list.

    Returns:
        list -- a list of bytes depicitng the rom
    """

    newlen = int(newlenmb) * 1024 * 1024
    if len(rom) > newlen:
        raise alttprException('ROM is already larger than {bytes}'.format(
            bytes=newlen
        ))
    diff = len(rom) - newlen
    if diff > 0:
        rom[newlen] = 0
    else:
        rom.extend(itertools.repeat(0, -diff))
        rom.append(0)
    return rom



import copy
import random
from .customizer import get_starting_equipment, BASE_CUSTOMIZER_PAYLOAD
from .misc import mergedicts

BASE_RANDOMIZER_PAYLOAD = {
    "glitches": "none",
    "item_placement": "advanced",
    "dungeon_items": "standard",
    "accessibility": "items",
    "goal": "ganon",
    "crystals": {
        "ganon": "7",
        "tower": "7"
    },
    "mode": "open",
    "entrances": "none",
    "hints": "off",
    "weapons": "randomized",
    "item": {
        "pool": "normal",
        "functionality": "normal"
    },
    "tournament": False,
    "spoilers": "on",
    "lang": "en",
    "enemizer": {
        "boss_shuffle": "none",
        "enemy_shuffle": "none",
        "enemy_damage": "default",
        "enemy_health": "default",
        "pot_shuffle": "off"
    },
    "allow_quickswap": False
}


def generate_random_settings(weights, tournament=True, spoilers="mystery"):
    subweight_name = get_random_option(
        {k: v['chance'] for (k, v) in weights.get('subweights', {}).items()})
    if subweight_name is not None:
        subweights = weights.get('subweights', {}).get(
            subweight_name, {}).get('weights', {})
        weights = {**weights, **subweights}

    # customizer isn't used until its used
    customizer = False

    # we need to figure out if entrance shuffle is a thing, since that tells us if we should even bother with rolling customizer things
    entrances = get_random_option(weights['entrance_shuffle'])

    # only roll customizer stuff if entrance shuffle isn't on, and we have a customizer section
    if entrances == "none" and weights.get('customizer', None):
        custom = {}
        eq = []
        pool = {}

        if 'eq' in weights['customizer']:
            for key in weights['customizer']['eq'].keys():
                value = get_random_option(weights['customizer']['eq'][key])
                if value:
                    eq += get_starting_equipment(key=key, value=value)
                    customizer = True

        if 'custom' in weights['customizer']:
            for key in weights['customizer']['custom'].keys():
                value = get_random_option(weights['customizer']['custom'][key])
                if value is not None:
                    custom[key] = value
                    customizer = True

        if 'pool' in weights['customizer']:
            for key in weights['customizer']['pool'].keys():
                value = get_random_option(weights['customizer']['pool'][key])
                if value is not None:
                    pool[key] = value
                    customizer = True

    if customizer:
        settings = copy.deepcopy(BASE_CUSTOMIZER_PAYLOAD)
    else:
        settings = copy.deepcopy(BASE_RANDOMIZER_PAYLOAD)

    settings["glitches"] = get_random_option(weights['glitches_required'])
    settings["item_placement"] = get_random_option(weights['item_placement'])
    settings["dungeon_items"] = get_random_option(weights['dungeon_items'])
    settings["accessibility"] = get_random_option(weights['accessibility'])
    settings["goal"] = get_random_option(weights['goals'])
    settings["crystals"]["ganon"] = get_random_option(weights['ganon_open'])
    settings["crystals"]["tower"] = get_random_option(weights['tower_open'])
    settings["mode"] = get_random_option(weights['world_state'])
    settings["hints"] = get_random_option(weights['hints'])
    settings["weapons"] = get_random_option(weights['weapons'])
    settings["item"]["pool"] = get_random_option(weights['item_pool'])
    settings["item"]["functionality"] = get_random_option(
        weights['item_functionality'])
    settings["tournament"] = tournament
    settings["spoilers"] = spoilers
    settings["enemizer"]["boss_shuffle"] = get_random_option(
        weights['boss_shuffle'])
    settings["enemizer"]["enemy_shuffle"] = get_random_option(
        weights['enemy_shuffle'])
    settings["enemizer"]["enemy_damage"] = get_random_option(
        weights['enemy_damage'])
    settings["enemizer"]["enemy_health"] = get_random_option(
        weights['enemy_health'])
    settings["enemizer"]["pot_shuffle"] = get_random_option(
        weights.get('pot_shuffle', 'off'))

    settings["allow_quickswap"] = get_random_option(
        weights.get('allow_quickswap', False))

    if customizer:
        # default to v31 prize packs
        settings['custom']['customPrizePacks'] = False

        # set custom settings that were rolled
        for key, value in custom.items():
            settings['custom'][key] = value

        # set custom item pool that was rolled
        for key, value in pool.items():
            settings['custom']['item']['count'][key] = value

        # apply custom starting equipment, and adjust the item pool accordingly
        if eq:
            # remove items from pool
            for item in eq:

                # remove flute if starting with activated flute
                # remove bottles as well
                # TODO: This code is terrible, it should be its own function
                if item == 'OcarinaActive':
                    item = 'OcarinaInactive'
                if item in ['Bottle', 'BottleWithRedPotion', 'BottleWithGreenPotion', 'BottleWithBluePotion', 'BottleWithBee', 'BottleWithGoldBee', 'BottleWithFairy']:
                    item = 'BottleWithRandom'

                settings['custom']['item']['count'][item] = settings['custom']['item']['count'].get(
                    item, 0) - 1 if settings['custom']['item']['count'].get(item, 0) > 0 else 0

            # re-add 3 heart containers as a baseline
            eq += ['BossHeartContainer'] * 3

            # update the eq section of the settings
            settings['eq'] = eq

        # if dark room navigation is enabled, then
        # oh and yes item.require.Lamp is mixed around for whatever reason
        # False = dark room navigation isn't required
        if settings['custom'].get('item.require.Lamp', False):
            settings['enemizer']['enemy_shuffle'] = 'none'
            settings['enemizer']['enemy_damage'] = 'default'
            settings['enemizer']['pot_shuffle'] = 'off'

        # set dungeon_items to standard if any region.wild* custom settings are present
        if any(key in ['region.wildKeys', 'region.wildBigKeys', 'region.wildCompasses', 'region.wildMaps'] for key in custom):
            settings['dungeon_items'] = 'standard'

        # set custom triforce hunt settings if TFH is the goal
        if settings['goal'] == 'triforce-hunt':
            if 'triforce-hunt' in weights['customizer']:
                min_difference = get_random_option(
                    weights['customizer']['triforce-hunt'].get('min_difference', 0))
                try:
                    goal_pieces = randval(
                        weights['customizer']['triforce-hunt']['goal'])
                except KeyError:
                    goal_pieces = 20

                try:
                    if isinstance(weights['customizer']['triforce-hunt']['pool'], list):
                        if weights['customizer']['triforce-hunt']['pool'][0] + min_difference < goal_pieces:
                            weights['customizer']['triforce-hunt']['pool'][0] = goal_pieces + \
                                min_difference
                    pool_pieces = randval(
                        weights['customizer']['triforce-hunt']['pool'])

                    # a final catchall
                    if pool_pieces < goal_pieces + min_difference:
                        pool_pieces = goal_pieces + min_difference
                except KeyError:
                    pool_pieces = 30
            else:
                goal_pieces = 20
                pool_pieces = 30

            settings['custom']['item.Goal.Required'] = goal_pieces
            settings['custom']['item']['count']['TriforcePiece'] = pool_pieces

        if settings['custom'].get('rom.timerMode', 'off') == 'countdown-ohko':
            if 'timed-ohko' in weights['customizer']:
                for clock in weights['customizer']['timed-ohko'].get('clock', {}):
                    settings['custom'][f'item.value.{clock}'] = randval(
                        weights['customizer']['timed-ohko']['clock'][clock].get(
                            'value', 0)
                    )
                    settings['custom']['item']['count'][clock] = randval(
                        weights['customizer']['timed-ohko']['clock'][clock].get(
                            'pool', 0)
                    )

                settings['custom']['rom.timerStart'] = randval(
                    weights['customizer']['timed-ohko'].get('timerStart', 0)
                )

            settings = dict(mergedicts(
                settings, weights['customizer']['timed-ohko'].get('forced_settings', {})))

        # fill in empty items in pool with FillItemPoolWith option, defaults to "Nothing"
        filler = weights.get('options', {}).get('FillItemPoolWith', 'Nothing')
        settings['custom']['item']['count'][filler] = settings['custom']['item']['count'].get(
            filler, 0) + max(0, 216 - sum(settings['custom']['item']['count'].values()))

        # deactivate a starting flute that's pre-activated, as it'll cause some really dumb rainstate scenarios
        if settings["mode"] == 'standard':
            settings['eq'] = [item if item !=
                              'OcarinaActive' else 'OcarinaInactive' for item in settings.get('eq', {})]

        # fix a bad interaction between pedestal/dungeons goals and prize.crossWorld
        if settings["goal"] in ['pedestal', 'dungeons']:
            settings['custom']['prize.crossWorld'] = True
    else:
        settings["entrances"] = entrances

    # If mc or mcs shuffle gets rolled, and its entrance, shift to either standard or full accordingly
    # mc and mcs are not supported by the entrance randomizer version currently used for v31.0.4.
    # If this changes, these statements will get removed.
    if settings.get('entrances', 'none') != 'none':
        if settings.get('dungeon_items', 'standard') == 'mc':
            settings['dungeon_items'] = 'standard'
        elif settings.get('dungeon_items', 'standard') == 'mcs':
            settings['dungeon_items'] = 'full'

    # This if statement is dedicated to the survivors of http://www.speedrunslive.com/races/result/#!/264658
    if settings['weapons'] not in ['vanilla', 'assured'] and settings['mode'] == 'standard' and (
            settings['enemizer']['enemy_shuffle'] != 'none'
            or settings['enemizer']['enemy_damage'] != 'default'
            or settings['enemizer']['enemy_health'] != 'default'):
        settings['weapons'] = 'assured'

    return settings, customizer


def conv(string):
    # first try to convert it to a integer

    if isinstance(string, str):
        try:
            return int(string)
        except ValueError:
            pass

        if string.lower() == "true":
            return True
        if string.lower() == "false":
            return False

    return string


def randval(optset):
    if isinstance(optset, list):
        return random.randint(optset[0], optset[1])
    else:
        return optset


def get_random_option(optset):
    if not optset:
        return None
    try:
        return random.choices(population=[conv(key) for key in list(optset.keys())], weights=list(optset.values()))[0] if isinstance(optset, dict) else conv(optset)
    except TypeError as err:
        raise TypeError("There is a non-numeric value as a weight.") from err
